from tqdm import tqdm

try:
	from .util_time import timed
except (ModuleNotFoundError,ImportError):
	from util_time import timed

class ANN():

	@timed
	def init_dense_ann(self, method='hnsw', space='cosinesimil', save=True, **kwargs):
		import nmslib
		self.dense_ann = nmslib.init(method=method, space=space)
		dense = self.dense
		dense = tqdm(dense, desc='ann_input')
		for i,point in enumerate(dense):
			self.dense_ann.addDataPoint(i,point)
		self.dense_ann.createIndex(kwargs, print_progress=True)
		if save:
			self.dense_ann.saveIndex(self.path+'ann_dense.bin',save_data=True)

	@timed
	def init_sparse_ann(self, method='hnsw', space='cosinesimil_sparse', save=True, **kwargs):
		import nmslib
		self.sparse_ann = nmslib.init(method=method, space=space, data_type=nmslib.DataType.SPARSE_VECTOR)
		sparse = self.sparse
		sparse = tqdm(sparse, desc='ann_input')
		for i,point in enumerate(sparse):
			self.sparse_ann.addDataPoint(i,point)		
		self.sparse_ann.createIndex(kwargs, print_progress=True)
		if save:
			self.sparse_ann.saveIndex(self.path+'ann_sparse.bin',save_data=True)

	def load_dense_ann(self):
		import nmslib
		self.dense_ann = nmslib.init(method='hnsw', space='cosinesimil')
		self.dense_ann.loadIndex(self.path+'ann_dense.bin',load_data=True)

	def load_sparse_ann(self):
		import nmslib
		self.sparse_ann = nmslib.init(method='hnsw', space='cosinesimil_sparse', data_type=nmslib.DataType.SPARSE_VECTOR)
		self.sparse_ann.loadIndex(self.path+'ann_sparse.bin',load_data=True)

	def sparse_ann_query(self, point, k=10):
		return self.sparse_ann.knnQuery(point, k)

	def dense_ann_query(self, point, k=10):
		return self.dense_ann.knnQuery(point, k)




def extract_cfg(cfg, prefix, sep='.'):
	out = {}
	for key,val in cfg.items():
		if not key.startswith(prefix): continue
		key = key[len(prefix)+len(sep):]
		if sep in key or not key: continue
		out[key] = val
	return out

if __name__=="__main__":
	cfg = {
		'a.1':'aaa',
		'a.2':'bbb',
		'a.x.1':'ccc',
		'a.x.2':'ddd',
		'b.x':'eee',
	}
	print(extract_cfg(cfg,'a.x'))
	


import os
import re
import pickle
from types import FunctionType

from tqdm import tqdm
import dill

from .model_params     import Params
from .model_meta       import Meta
from .model_sentencer  import Sentencer
from .model_phraser    import Phraser
from .model_dictionary import Dictionary
from .model_tfidf      import TFIDF
from .model_lsi        import LSI
from .model_ann        import ANN
from .model_subword    import Subword

from .util_time import timed

split_sentences_re = re.compile('[.?!]+ (?=[A-Z])')
split_tokens_re = re.compile('[\s.,;!?()\[\]]+')

class Model(
		Params,
		Meta,
		Sentencer,
		Phraser,
		Dictionary,
		TFIDF,
		LSI,
		ANN,
		Subword
	):
	
	def __init__(self, path):
		Params.__init__(self)
		self.set_params(path=path)
		# create model directory if not exists
		model_dir = os.path.dirname(path)
		if not os.path.exists(model_dir):
			os.makedirs(model_dir)
	
	@property
	def path(self):
		return self.params['path']
	
	#@timed
	def find_sparse(self, text, k=10):
		sparse = self.text_to_sparse(text)
		if sparse:
			i_list,d_list = self.sparse_ann_query(sparse,k)
			for i,d in zip(i_list,d_list):
				m = self.meta[i]
				yield i,d,m

	#@timed
	def find_dense(self, text, k=10):
		dense = self.text_to_dense(text)
		if dense:
			i_list,d_list = self.dense_ann_query(dense,k)
			for i,d in zip(i_list,d_list):
				m = self.meta[i]
				yield i,d,m

	def text_to_sparse(self, text):
		phrased = self.text_to_phrased(text)
		bow = self.dictionary.doc2bow(phrased)
		return self.tfidf[bow]
	
	def text_to_dense(self, text):
		sparse = self.text_to_sparse(text)
		return self.sparse_to_dense(sparse)

	@staticmethod
	def text_to_sentences(text):
		return split_sentences_re.split(text)

	@staticmethod
	def text_to_tokens(text):
		tokens = split_tokens_re.split(text)
		return [t.lower() for t in tokens]

	@staticmethod
	def doc_to_text(doc):
		values = [x for x in doc.values() if type(x) is str]
		return '\n\n'.join(values)

	def explain(self,id):
		phrased = self.phrased[id]
		ids = [self.dictionary.token2id.get(p,-1) for p in phrased]
		return [p for p,i in zip(phrased,ids) if i>=0]

	# TODO opcja do ladowania wszystkiego co jest zapisane
	# TODO analogiczna funkcja init ala sklearn.model
	@timed
	def load(self, components=[]):
		# TODO wykrywanie ktore komponenty wczytac
		if 'fun'        in components: self.load_fun()
		#
		if 'meta'       in components: self.load_meta()
		if 'sentencer'  in components: self.load_sentencer()
		if 'phraser'    in components: self.load_phraser()
		if 'dictionary' in components: self.load_dictionary()
		if 'tfidf'      in components: self.load_tfidf()
		if 'lsi'        in components: self.load_lsi()
		if 'sparse_ann' in components: self.load_sparse_ann()
		if 'dense_ann'  in components: self.load_dense_ann()
		#
		if 'phrased'    in components: self.load_phrased()
		if 'bow'        in components: self.load_bow()
		if 'sparse'     in components: self.load_sparse()
		if 'dense'      in components: self.load_dense()
		if 'inverted'   in components: self.load_inverted()
		#
		return self

	# EXPERIMENTAL
	def get_doc(self, i):
		"""get single document (paragraph) by index"""
		meta = self.meta[i]
		return self.get_doc_by_meta(meta)

	# EXPERIMENTAL
	def init_fun(self):
		fun_dict = {}
		for name in dir(self):
			if name[0]=='_':continue
			fun = getattr(self, name)
			if type(fun) is FunctionType:
				fun_dict[name] = fun
		dill.dump(fun_dict, open(self.path+'fun.dill','wb'), recurse=True)

	# EXPERIMENTAL
	def load_fun(self):
		fun_dict = dill.load(open(self.path+'fun.dill','rb'))
		for name,fun in fun_dict.items():
			setattr(self, name, fun)

# ------------------------------------------------------------------------------

if __name__=="__main__":
	pass



from gensim.models import FastText

try:
	from .util_time import timed
except (ModuleNotFoundError,ImportError):
	from util_time import timed

from time import time

class Subword():

	@timed
	def init_subword(self, size, window=3, min_count=1, epochs=10, workers=None):
		t0 = time()
		_workers = workers or self.params.get('subword__workers') or self.params.get('workers',1)
		corpus_file = self.params.get('subword__corpus_file') or self.params.get('corpus_file') or 'sentences.txt.gz'
		self.subword = FastText(size=size, window=window, min_count=min_count, workers=_workers)
		self.subword.build_vocab(corpus_file=self.path+corpus_file)
		print(f'corpus_count: {self.subword.corpus_count}')
		print(f'sen_cnt: {self.sentences_cnt}')
		print(f'build_vocab {time()-t0:.01f} seconds') # XXX
		self.subword.train(
				corpus_file=self.path+corpus_file,
				#total_examples=self.subword.corpus_count,
				total_words=self.tokens_cnt,
				epochs=epochs
			)
		self.subword.save(self.path+'subword')
	
	def load_subword(self):
		self.subword = FastText.load(self.path+'subword')
		return self


from gensim.models.phrases import Phrases
from gensim.models.phrases import Phraser as GensimPhraser
from gensim.models.word2vec import LineSentence
from tqdm import tqdm

from collections import Counter
import multiprocessing as mp
import pickle
import math
import gzip
import os

try:
	from .util_time import timed
	from .sorbet import sorbet
except (ModuleNotFoundError,ImportError):
	from util_time import timed
	from sorbet import sorbet

# ---[ MODEL ]------------------------------------------------------------------

class phraser:

	def __getitem__v1(self, tokens, dlm='_'):
		if not tokens: return []
		out = [tokens[0]]
		for i in range(1,len(tokens)):
			if (tokens[i-1],tokens[i]) in self.pairs:
				out[-1] = out[-1]+dlm+tokens[i]
			else:
				out.append(tokens[i])
		return out

	def __getitem__(self, tokens, dlm='_'):
		if not tokens: return []
		out = [tokens[0]]
		for a,b in zip(tokens,tokens[1:]):
			if (a,b) in self.pairs:
				out[-1] = out[-1] +dlm+ b
			else:
				out.append(b)
		return out


class Phraser():

	@timed
	def init_phraser(self, components=False, **kwargs):
		sentences = LineSentence(self.path+'sentences.txt.gz')
		phrases = Phrases(sentences, **kwargs)
		self.phraser = GensimPhraser(phrases)
		self.phraser.components = components
		self.phraser.save(self.path+'phraser.pkl')
		del phrases
	
	@timed
	def init_phraser_mk2(self):
		self.phraser = phraser()
		min_freq = 10 # TODO params
		min_npmi = 0.8 # TODO params
		cf = Counter()
		pairs = []
		tok_cnt = 0
		# TODO mozliwosc wykonania na raty (aby oszczedzac RAM)
		for _,sen in self.all_sentences('sentences.txt.gz',desc='phraser_mk2'):
			tokens = sen.split(' ')
			tok_cnt += len(tokens)
			pairs.extend(zip(tokens,tokens[1:]))
			cf.update(tokens)
		pf = Counter(pairs)
		#
		log = math.log
		pairs = set()
		npmis = {} # diag
		pmis = {} # diag
		pfs = {} # diag
		for (a,b),fp in pf.items():
			if fp >= min_freq:
				pa = cf[a] / tok_cnt
				pb = cf[b] / tok_cnt
				pp = fp / tok_cnt
				expected = pa * pb * tok_cnt
				lift = fp / expected
				pmi = log( pp / (pa*pb) ) # point-wise mutual information
				npmi = pmi / -log(pp) # normalized pmi
				if npmi >= min_npmi:
					pairs.add((a,b))
					npmis[(a,b)] = npmi
					pmis[(a,b)] = pmi
					pfs[(a,b)] = fp
					#print(f'pair {npmi:.03f} {a} {b} {fp}') # XXX
		self.phraser.pairs = pairs
		self.phraser.npmis = npmis
		self.phraser.pmis = pmis
		self.phraser.pfs = pfs
		pickle.dump(self.phraser, open(self.path+'phraser_mk2.pkl','wb'))
	
	
	def load_phraser_mk2(self):
		self.phraser = pickle.load(open(self.path+'phraser_mk2.pkl','rb'))

	@timed
	def init_phrased_mk2(self):
		gzip_level = self.params.get('phrased__gzip_level',1) 
		f = gzip.open(self.path+'phrased.txt.gz', 'wt',
		              encoding='utf8', compresslevel=gzip_level)
		for _,sen in self.all_sentences('sentences.txt.gz', desc='phrased_mk2'):
			if not sen:
				f.write('\n')
				continue
			tokens = sen.split(' ')
			phrased = ' '.join(self.phraser[tokens])
			f.write(phrased)
			f.write('\n')
		f.close()
	
	def skip_phraser(self):
		self.phraser = None
		open(self.path+'phraser.pkl','wb') # empty file -> skip_phraser

	def load_phraser(self):
		path = self.path+'phraser.pkl'
		if os.path.getsize(path):
			self.phraser = GensimPhraser.load(path)
		else:
			self.phraser = None
	
	def doc_to_phrased(self, doc):
		text = self.doc_to_text(doc)
		yield from self.text_to_phrased(text)
	
	def text_to_phrased(self, text):
		if self.phraser:
			components = self.phraser.components
			sentences = self.text_to_sentences(text)
			for sentence in sentences:
				tokens = self.text_to_tokens(sentence)
				phrased = self.phraser[tokens]
				yield from phrased
				if components:
					yield from set(tokens)-set(phrased)
		else:
			yield from self.text_to_tokens(text)
	
	# TODO jako plik txt.gz
	@timed
	def init_phrased(self, storage='disk'):
		documents = self.doc_iter()
		phrased = (list(self.doc_to_phrased(doc)) for doc in documents)
		phrased = tqdm(phrased, desc='phrased', total=len(self.meta)) # progress bar
		self.phrased = sorbet(self.path+'phrased', kind=storage).dump(phrased)

	@timed
	def init_phrased_mp(self, workers=4, chunksize=100):
		s = sorbet(self.path+'phrased').new()
		doc_id_iter = range(len(self.meta))
		doc_id_iter = tqdm(doc_id_iter,'phrased',len(self.meta))
		with mp.Pool(
					workers,
					init_phrased_worker,
					[
						self.path,
						self.get_doc_by_meta,
						self.doc_to_text,
						self.text_to_sentences,
						self.text_to_tokens,
						self.phraser==None
					]
				) as pool:
			phrased = pool.imap(phrased_worker, doc_id_iter, chunksize)
			for p in phrased:
				s.append(list(p))
		self.phrased = s.save()
	
	def skip_phrased(self):
		doc_to_text = self.doc_to_text
		text_to_tokens = self.text_to_tokens
		doc_iter = self.doc_iter
		class __phrased:
			def __iter__(self):
				return (list(text_to_tokens(doc_to_text(doc))) for doc in doc_iter())
			def delete(self): pass
		self.phrased = __phrased()
	
	def load_phrased(self, storage='disk'):
		self.phrased = sorbet(self.path+'phrased', kind=storage).load()

# ---[ MULTIPROCESSING ]--------------------------------------------------------

def init_phrased_worker(*args):
	global model
	model = Phraser()
	model.path              = args[0]
	model.get_doc_by_meta   = args[1]
	model.doc_to_text       = args[2]
	model.text_to_sentences = args[3]
	model.text_to_tokens    = args[4]
	model.skip_phraser      = args[5]
	model.meta = sorbet(model.path+'meta').load()
	if not model.skip_phraser:
		model.load_phraser()
	
def phrased_worker(doc_id):
	meta = model.meta[doc_id]
	doc = model.get_doc_by_meta(meta)
	text = model.doc_to_text(doc)
	if model.skip_phraser:
		out = model.text_to_tokens(text)
	else:
		out = []
		components = model.phraser.components
		for sen in model.text_to_sentences(text):
			tokens = model.text_to_tokens(sen)
			phrased = model.phraser[tokens]
			out.extend(phrased)
			if components:
				out.extend(set(tokens)-set(phrased)) 
	return out

# ---[ DEBUG ]-------------------------------------------------------------------

def count_digits(tokens):
	x = 0
	for token in tokens:
		for letter in token:
			if letter in "0123456789":
				x += 1
	return x

if __name__=="__main__":
	model = Phraser()
	model.path = 'model_100/'
	model.load_phraser()
	model.load_phrased()
	#
	from pprint import pprint
	#pprint(model.phrased[0])
	#
	#
	from time import time
	t0=time()
	total = 0
	aaa = {}
	aaa['total'] = 0
	if 0: # 4000 -> 35s
		for tokens in model.phrased:
			x = 0
			for token in tokens:
				for letter in token:
					if letter in "0123456789":
						x += 1
			total += x
	if 1: # 4000 -> 2p:22s
		import multiprocessing as mp
		from tqdm import tqdm
		pool = mp.Pool(processes=2)
		def agg(x):
			aaa['total'] += 1
		#results = pool.map(count_digits, tqdm(model.phrased), 100)
		#results = pool.imap_unordered(count_digits, tqdm(model.phrased), 100)
		results = pool.map_async(count_digits, tqdm(model.phrased), 100, agg)
		for x in results.get():
			total += x
	print(total)
	print(aaa['total'])
	print(f"done in {time()-t0:.02f} seconds")



from tqdm import tqdm
import pickle
from hashlib import md5
from collections import Counter
import multiprocessing as mp
import gzip

try:
	from .util_time import timed
	from .sorbet import sorbet
except (ModuleNotFoundError,ImportError):
	from util_time import timed
	from sorbet import sorbet


# def my_hash(text):
	# """64bit hash from text"""
	# return int(md5(text.lower().encode()).hexdigest()[:16],16)

class sentencer: pass

class Sentencer:

	@timed
	def init_sentencer(self, workers=None):
		"""store line sentences after tokenization
		
		empty sentences are omitted
		an empty line marks the end of the document
		"""
		self.sentencer = sentencer()
		_workers = workers or self.params.get('sentences__workers') or self.params.get('workers',1)
		if _workers>1:
			self._init_sentences_mp(_workers)
		else:
			self._init_sentences_sp()
		pickle.dump(self.sentencer, open(self.path+'sentencer.pkl','wb'))

	def load_sentencer(self):
		self.offset = sorbet(self.path+'offset').load()
		self.sentencer = pickle.load(open(self.path+'sentencer.pkl','rb'))
		return self

	def all_sentences(self, corpus_file=None, desc='sentences'):
		_corpus_file = corpus_file or self.params.get('corpus_file') or 'sentences.txt.gz'
		f = gzip.open(self.path+_corpus_file, 'rt', encoding='utf8')
		f = tqdm(f, desc, total=self.sentencer.sentences_cnt)
		doc_id = 0
		for line in f:
			sen = line.rstrip()
			yield doc_id,sen 
			if not sen: # end of document
				doc_id += 1
		f.close()

	def doc_sentences(self, doc_id, corpus_file=None):
		_corpus_file = corpus_file or self.params.get('corpus_file') or 'sentences.txt.gz'
		try:
			pos = self.offset[doc_id]
		except IndexError:
			raise Exception(f'IndexError: offset[{doc_id}] len(offset)={len(self.offset)}')
		f = gzip.open(self.path+_corpus_file, 'rt', encoding='utf8')
		f.seek(pos)
		for line in f:
			sen = line.rstrip()
			if not sen: break # end of document
			yield sen
		f.close()

	def raw_sentences(self, desc='raw_sentences'):
		"""iterate over doc_id,sentence before tokenization"""
		doc_iter = self.doc_iter()
		doc_iter = tqdm(doc_iter, desc=desc, total=len(self.meta))
		for doc_id,doc in enumerate(doc_iter):
			text = self.doc_to_text(doc)
			sentences = self.text_to_sentences(text)
			for sen in sentences:
				yield doc_id,sen
	

	def _init_sentences_sp(self):
		offset = sorbet(self.path+'offset').new()
		gzip_level = self.params.get('sentences__gzip_level',1) 
		f = gzip.open(self.path+'sentences.txt.gz', 'wt',
		              encoding='utf8', compresslevel=gzip_level)
		text_to_tokens = self.text_to_tokens
		sentences = self.raw_sentences(desc='sentences')
		pos = 0
		sen_cnt = 0
		tok_cnt = 0
		prev_doc_id = None
		for doc_id,sentence in sentences:
			# end of document
			if doc_id!=prev_doc_id and prev_doc_id is not None:
				f.write('\n')
				offset.append(pos)
				pos = f.tell()
			tokens = text_to_tokens(sentence)
			tok_cnt += len(tokens)
			if not tokens: continue # skip empty sentences
			text = ' '.join(tokens)
			f.write(text)
			f.write('\n')
			prev_doc_id = doc_id
			sen_cnt += 1
		# end of the last document
		f.write('\n')
		offset.append(pos)
		
		f.close()
		offset.save()
		self.offset = offset
		self.sentencer.sentences_cnt = sen_cnt
		self.sentencer.tokens_cnt = tok_cnt
	
	
	def _init_sentences_mp(self, workers=2):
		chunk = self.params.get('sentences__chunk',100)
		offset = sorbet(self.path+'offset').new()
		doc_id_iter = range(len(self.meta))
		doc_id_iter = tqdm(doc_id_iter, 'sentences', len(self.meta))
		compresslevel = 1 # TODO cfg
		f = gzip.open(self.path+'sentences.txt.gz', 'wt',
		              encoding='utf8', compresslevel=compresslevel)
		with mp.Pool(
					workers,
					init_sentences_worker,
					[
						self.path,
						self.get_doc_by_meta,
						self.doc_to_text,
						self.text_to_sentences,
						self.text_to_tokens,
					]
				) as pool:
			doc_sen_iter = pool.imap(sentences_worker, doc_id_iter, chunk)
			sen_cnt = 0
			tok_cnt = 0 # TODO
			pos = 0
			for doc_sen in doc_sen_iter:
				offset.append(pos)
				if doc_sen:
					f.write('\n'.join(doc_sen))
					f.write('\n\n')
				else: # empty document
					f.write('\n')
				start_pos = pos
				pos = f.tell()
				length = pos-start_pos
				# TODO store start_pos+length
				sen_cnt += len(doc_sen)
		f.close()
		offset.save()
		self.offset = offset
		self.sentencer.sentences_cnt = sen_cnt
		self.sentencer.tokens_cnt = tok_cnt # TODO



def init_sentences_worker(*args):
	global model
	model = Sentencer()
	model.path              = args[0]
	model.get_doc_by_meta   = args[1]
	model.doc_to_text       = args[2]
	model.text_to_sentences = args[3]
	model.text_to_tokens    = args[4]
	model.meta = sorbet(model.path+'meta').load()

def sentences_worker(doc_id):
	meta = model.meta[doc_id]
	doc = model.get_doc_by_meta(meta)
	text = model.doc_to_text(doc)
	out = []
	tok_cnt = 0
	for sentence in model.text_to_sentences(text):
		tokens = model.text_to_tokens(sentence)
		tok_cnt += len(tokens)
		if not tokens: continue # skip empty sentences
		out += [' '.join(tokens)]
	return out

if __name__ == "__main__":
	import data
	model = Sentencer()
	model.path = 'model_100/'
	model.load_sentencer()


from gensim.models import TfidfModel
from tqdm import tqdm
import multiprocessing as mp
import re
from array import array

try:
	from .util_time import timed
	from .sorbet import sorbet
except (ModuleNotFoundError,ImportError):
	from util_time import timed
	from sorbet import sorbet

# ---[ MODEL ]------------------------------------------------------------------

class TFIDF():
	
	@timed
	def init_tfidf(self, **kwargs):
		self.tfidf = TfidfModel(dictionary=self.dictionary, **kwargs)
		self.tfidf.save(self.path+'tfidf.pkl')
	
	def load_tfidf(self):
		self.tfidf = TfidfModel.load(self.path+'tfidf.pkl')


	def init_sparse(self, workers=None, storage=None):
		_workers = workers or self.params.get('sparse__workers') or self.params.get('workers',1)
		_storage = storage or self.params.get('sparse__storage') or self.params.get('storage','disk')
		if _workers>1:
			self._init_sparse_mp(workers=_workers, storage=_storage)
		else:
			self._init_sparse_sp(storage=_storage)

	def _init_sparse_sp(self, storage):
		self.sparse = sorbet(self.path+'sparse', kind=storage).new()
		doc = []
		for doc_id,sen in self.all_sentences(desc='sparse'):
			if not sen: # end of document:
				bow = self.dictionary.doc2bow(doc) or [(0,1)] # 0 -> '<PAD>'
				sparse = self.tfidf[bow]
				a = array('f')
				for i_v in sparse:
					a.extend(i_v)
				self.sparse.append(a)
				doc = []
			else:
				tokens = sen.split(' ')
				doc.extend(tokens)
		self.sparse.save()


	def load_sparse(self, storage='disk'):
		self.sparse = sorbet(self.path+'sparse', kind=storage).load()
	
	def _init_sparse_mp(self, workers, storage):
		chunksize = self.params.get('sparse__chunksize',10)
		s = sorbet(self.path+'sparse').new()
		id_iter = range(len(self.meta))
		id_iter = tqdm(id_iter,'sparse',len(self.meta))
		with mp.Pool(
					workers,
					init_sparse_worker,
					[
						self.path,
					]
				) as pool:
			sparse = pool.imap(sparse_worker, id_iter, chunksize)
			for a in sparse:
				#print('mp_sparse',a)
				s.append(a)
		self.sparse = s.save()
	
	# TODO raname dfs_query ???
	def tfidf_query(self, query, exclude=None):
		dfs = []
		token_ids,tokens = self.dictionary_query(query,exclude)
		for i in token_ids:
			df = self.tfidf.dfs[i]
			dfs += [df]
		# sort
		results = list(sorted(zip(token_ids,tokens,dfs),key=lambda x:-x[2]))
		token_ids = [x[0] for x in results]
		tokens = [x[1] for x in results]
		dfs = [x[2] for x in results]
		return token_ids,tokens,dfs
			

# ---[ MULTIPROCESSING ]--------------------------------------------------------

def init_sparse_worker(*args):
	global model
	model_path = args[0]
	model = TFIDF()
	model.path = model_path
	model.bow = sorbet(model.path+'bow').load()
	model.load_tfidf()


def sparse_worker(doc_id):
	bow = model.bow[doc_id]
	sparse = model.tfidf[bow]
	a = array('f')
	for i_v in sparse:
		a.extend(i_v)
	return a


# tragiczna wydajnosc 
def init_sparse_worker_mk2(*args):
	global model
	from .model_dictionary import Dictionary
	from .model_sentencer import Sentencer
	class Model(TFIDF,Dictionary,Sentencer): pass
	model_path = args[0]
	model = Model()
	model.path = model_path
	model.load_tfidf()
	model.load_dictionary()
	model.load_sentencer()


# tragiczna wydajnosc 
def sparse_worker_mk2(doc_id):
	tokens = []
	for sen in model.doc_sentences(doc_id):
		tokens.extend(sen.split(' '))
	bow = model.dictionary.doc2bow(tokens) or [(0,1)] # 0 -> '<PAD>'
	sparse = model.tfidf[bow]
	a = array('f')
	for i_v in sparse:
		a.extend(i_v)
	return a


# ---[ DEBUG ]------------------------------------------------------------------

if __name__=="__main__":
	model = TFIDF()
	model.path = '../model_100/'
	model.load_tfidf()
	tfidf = model.tfidf
	#
	from pprint import pprint
	from itertools import islice
	#
	from model_dictionary import Dictionary
	model2 = Dictionary()
	model2.path = model.path
	model2.load_dictionary()
	d = model2.dictionary
	data = ((i,cnt,tfidf.idfs[i],d.get(i,'???')) for i,cnt in tfidf.dfs.items())
	data = sorted(data,key=lambda x:x[1],reverse=True)
	for i,cnt,idf,token in islice(data,40):
		print(i,cnt,
			round(idf,1),
			token,
			sep='\t')



"""Sorbet - frozen sequence storage engine

Sorbet is fast - on a "standard" laptop it achieves:
- 200k writes per second,
- 300k reads per second.
"""

# ------------------------------------------------------------------------------

def sorbet(path, kind='disk'):
	if kind=='disk':
		return sorbet_on_disk(path)
	elif kind=='mem':
		return sorbet_in_mem(path)
	elif kind=='mem_only':
		return sorbet_in_mem_only(path)
	else:
		raise Exception(f"Invalid sorbet kind: '{kind}'")
		
# ------------------------------------------------------------------------------

from pickle import dump,load,HIGHEST_PROTOCOL
import os

# TODO parallel scan -> zwykly nie ma sensu bo jest funkcja filter
# TODO multiproc scan -> (path,index[lo:hi]->dict)

class sorbet_on_disk:
	"""frozen sequence storage engine"""
	
	def __init__(self, path, mp_pool=None):
		"""configure prefix of sorbet files"""
		self.path = path
		self.mp_pool = mp_pool
		self.protocol = HIGHEST_PROTOCOL
	
	def new(self):
		"""create new storage for appending data"""
		self.f = open(f"{self.path}.data",'wb')
		self.index = []
		return self
	
	def save(self):
		"""save data on disk; removes ability to append new data"""
		dump(self.index, open(f"{self.path}.index",'wb'), protocol=self.protocol)
		self.f.close()
		self.f = open(f"{self.path}.data",'rb')
		print(f"size of {self.path}.data - {os.path.getsize(f'{self.path}.data')/1_000_000:.01f} MB") # XXX
		return self
	
	def load(self):
		"""read data from disk"""
		self.f = open(f"{self.path}.data",'rb')
		self.index = load(open(f"{self.path}.index",'rb'))
		return self
	
	def dump(self, data):
		"""save data on disk"""
		self.new()
		f = self.f
		index = self.index
		p = self.protocol
		for val in data:
			index.append( f.tell() )
			dump(val, f, protocol=p)
		f.close()
		dump(index, open(f"{self.path}.index",'wb'), protocol=self.protocol)
		self.f = open(f"{self.path}.data",'rb')
		print(f"size of {self.path}.data - {os.path.getsize(f'{self.path}.data')/1_000_000:.01f} MB") # XXX
		return self
	
	def append(self, val):
		"""append value (of any type) to the dataset"""
		self.index.append( self.f.tell() )
		dump(val, self.f, protocol=self.protocol)
	
	def delete(self):
		self.f.close()
		os.remove(f"{self.path}.index")
		os.remove(f"{self.path}.data")
	
	def __getitem__(self, key):
		"""return value for given key or iterator for given slice"""
		if type(key) is slice:
			return self.__getslice__(key)
		else:
			pos = self.index[key]
			self.f.seek(pos)
			return load(self.f)
	
	def __getslice__(self, key):
		"""return iterator over given slice"""
		start,stop,step = key.indices(len(self))
		for i in range(start,stop,step):
			yield self[i]
	
	def __len__(self):
		"""return number of items"""
		return len(self.index)
	
	# experimental
	def imap(self, fun):
		yield from self.mp_pool.imap(fun, self)

# ------------------------------------------------------------------------------

class sorbet_in_mem:
	"""frozen sequence storage engine"""
	
	def __init__(self, path):
		"""configure prefix of sorbet files"""
		self.path = path
		self.protocol = HIGHEST_PROTOCOL
	
	def new(self):
		"""create new storage for appending data"""
		self.f = open(f"{self.path}.data",'wb')
		self.data = []
		return self
	
	def save(self):
		"""save data on disk; removes ability to append new data"""
		dump(self.data, self.f, protocol=self.protocol)
		self.f.close()
		print(f"size of {self.path}.data - {os.path.getsize(f'{self.path}.data')/1_000_000:.01f} MB") # XXX
		return self
	
	def load(self):
		"""read data from disk"""
		self.f = open(f"{self.path}.data",'rb')
		self.data = load(open(f"{self.path}.data",'rb'))
		return self
	
	def dump(self, data):
		"""save data on disk"""
		self.new()
		self.data = list(data)
		f = self.f
		p = self.protocol
		dump(self.data, f, protocol=p)
		f.close()
		print(f"size of {self.path}.data - {os.path.getsize(f'{self.path}.data')/1_000_000:.01f} MB") # XXX
		return self
	
	def append(self, val):
		"""append value (of any type) to the dataset"""
		self.data.append(val)
	
	def delete(self):
		del self.data
		os.remove(f"{self.path}.data")
	
	def __getitem__(self, key):
		"""return value for given key or iterator for given slice"""
		return self.data[key]
		
	def __len__(self):
		"""return number of items"""
		return len(self.data)

# ------------------------------------------------------------------------------

class sorbet_in_mem_only:
	"""frozen sequence storage engine"""
	
	def __init__(self, path):
		self.data = []
	
	def new(self):
		return self
	
	def save(self):
		return self
	
	def load(self):
		assert False
	
	def dump(self, data):
		self.data = list(data)
		return self
	
	def append(self, val):
		self.data.append(val)
	
	def delete(self):
		del self.data
	
	def __getitem__(self, key):
		return self.data[key]
		
	def __len__(self):
		return len(self.data)

# ---[ TEST ]-------------------------------------------------------------------

if __name__=="__main__":
	from time import time
	import sys
	label = sys.argv[0][:-3]
	path = f'../data/{label}'
	
	N = 1_100_000
	kind = 'disk'
	if 1:
		data = ({'a':i,'aa':i*10} for i in range(N))
		t0=time()
		db = sorbet(path,kind).dump(data)
		print("write time:",f"{time()-t0:.02f}s",f'{N/(time()-t0):.0f} items/s')
	if 1:
		db = sorbet(path,kind).load()
		t0=time()
		for i in range(N):
			db[i]
		print("read time:",f"{time()-t0:.02f}s",f'{N/(time()-t0):.0f} items/s')
	exit()
	
	print()
	print(list(db[:3]))
	print()
	for x in list(db[:3]):
		print(x)
	print()
	print(db[10])
	print('\nfilter')
	for x in filter(lambda x:x['a']%10000==0, db):
		print(x)
	print('\nfilter - slice')
	for x in filter(lambda x:x['a']%10000==0, db[50000:]):
		print(x)


	



from tqdm import tqdm

try:
	from .util_time import timed
	from .sorbet import sorbet
except (ModuleNotFoundError,ImportError):
	from util_time import timed
	from sorbet import sorbet


class Meta():

	@timed
	def init_meta(self, storage='disk'):
		self.meta = sorbet(self.path+'meta', kind=storage).new()
		get_meta = self.get_meta
		documents = self.doc_iter()
		documents = tqdm(documents, desc='meta')
		for id,doc in enumerate(documents):
			m = get_meta(id,doc)
			self.meta.append(m)
		self.meta.save()
	
	def load_meta(self, storage='disk'):
		self.meta = sorbet(self.path+'meta', kind=storage).load()



from docstr_coverage.coverage import get_docstring_coverage
from pprint import pprint
from pathlib import Path

path_list = [str(p) for p in Path('.').glob('*.py')]
pprint(get_docstring_coverage(path_list))





from gensim.models import LsiModel
from gensim.matutils import sparse2full
from tqdm import tqdm
from array import array
import multiprocessing as mp
import sys

try:
	from .util_time import timed
	from .sorbet import sorbet
except (ModuleNotFoundError,ImportError):
	from util_time import timed
	from sorbet import sorbet

# ---[ MODEL ]------------------------------------------------------------------

class LSI():
	
	@timed
	def init_lsi(self, **kwargs):		
		# handle onepass=False
		model_sparse = self.sparse
		class __sparse():
			def __iter__(self):
				for a in model_sparse:
					yield [(int(a[i]),a[i+1]) for i in range(0,len(a),2)]

		self.lsi = LsiModel(__sparse(), **kwargs)
		self.lsi.save(self.path+'lsi.pkl')

	
	def load_lsi(self):
		self.lsi = LsiModel.load(self.path+'lsi.pkl')

	def load_dense(self, storage='disk'):
		self.dense = sorbet(self.path+'dense', kind=storage).load()
	
	def sparse_to_dense(self, sparse):
		dense = self.lsi[sparse]
		dense = sparse2full(dense, self.lsi.num_topics)
		dense = array('f',dense)
		return dense

	@timed
	def init_dense(self, storage=None, workers=None):
		_workers = workers or self.params.get('dense__workers') or self.params.get('workers',1)
		_storage = storage or self.params.get('dense__storage') or self.params.get('storage','disk')
		if _workers>1:
			self._init_dense_mp(workers=_workers, storage=_storage)
		else:
			self._init_dense_sp(storage=_storage)

	def _init_dense_sp(self, storage='disk'):
		self.dense = sorbet(self.path+'dense', kind=storage).new()
		for a in self.sparse:
			sparse = [(int(a[i]),a[i+1]) for i in range(0,len(a),2)]
			dense = self.sparse_to_dense(sparse)
			self.dense.append(dense)
		self.dense.save()

	def _init_dense_mp(self, workers, storage):
		chunksize = self.params.get('dense__chunksize',10)
		s = sorbet(self.path+'dense').new()
		id_iter = range(len(self.meta))
		id_iter = tqdm(id_iter,'dense',len(self.meta))
		with mp.Pool(
					workers,
					init_dense_worker,
					[
						self.path,
					]
				) as pool:
			dense = pool.imap(dense_worker, id_iter, chunksize)
			for d in dense:
				s.append(d)
		self.dense = s.save()

# ---[ MULTIPROCESSING ]--------------------------------------------------------

def init_dense_worker(*args):
	global model
	model_path = args[0]
	model = LSI()
	model.path = model_path
	model.sparse = sorbet(model.path+'sparse').load()
	model.load_lsi()

def dense_worker(doc_id):
	a = model.sparse[doc_id]
	sparse = [(int(a[i]),a[i+1]) for i in range(0,len(a),2)]
	dense = model.lsi[sparse]
	dense = sparse2full(dense, model.lsi.num_topics)
	dense = array('f',dense)
	return dense


# ---[ DEBUG ]------------------------------------------------------------------

if __name__=="__main__":
	model = LSI()
	model.path = 'model_100/'
	model.load_lsi()
	model.load_dense()
	#
	print(model.dense[0])




from time import time
import sys

# TODO zmienna srodowiskowa sterujaca czy ma byc output i gdzie (out/err)
# TODO prefix nazwy funkcji -> nazwa pliku (bez .py)
def timed(fun):
	def wrapped(*args,**kwargs):
		t0 = time()
		out = fun(*args,**kwargs)
		print(f'{fun.__name__} done in {time()-t0:.02f} seconds',file=sys.stderr)
		return out
	return wrapped



from setuptools import setup, find_packages

setup(
	name="inverness",
	version="0.1.2",
	packages=find_packages(),
	
	install_requires=["gensim","nmslib","tqdm","dill"],
	
	author="Maciej Obarski",
	author_email="mobarski@gmail.com",
	description=\
		"Natural Language Processing framework "\
		"built on top of gensim and nmslib.",
	keywords="NLP",
	classifiers=[
		"License :: OSI Approved :: MIT License"
	]
)




class Params:

	def __init__(self):
		self.params = {}

	def set_params(self,**params):
		for k in params:
			self.params[k] = params[k]
	
	def get_params(self, deep=True):
		return self.params if deep else self.extract_params()

	# NOT USED ???
	def extract_params(self, component='', sep='__'):
		out = {}
		for key,val in self.params.items():
			if component:
				if not key.startswith(component): continue
				key = key[len(component)+len(sep):]
			if sep in key or not key: continue
			out[key] = val
		return out


class Model(Params):
	
	def __init__(self):
		print("hello")
		Params.__init__(self)

if __name__ == "__main__":
	p = Model()
	p.set_params(a=1,b=2,c__x=3,c__y=4,d__x=5,d__y__q=6,d__y__w=7)
	print(p.extract_params(''))



from gensim.corpora.dictionary import Dictionary as GensimDictionary
from tqdm import tqdm
import re
import pickle
from array import array

try:
	from .util_time import timed
	from .sorbet import sorbet
except (ModuleNotFoundError,ImportError):
	from util_time import timed
	from sorbet import sorbet

# TODO dodatkowa metoda - zamiast doc->bow doc->ids

# ---[ MODEL ]------------------------------------------------------------------

class Dictionary():
	
	@timed
	def init_dictionary(self, save=True):
		import gzip
		from collections import Counter
		corpus_file = self.params.get('dictionary__corpus_file') or self.params.get('corpus_file') or 'sentences.txt.gz'
		doc_id = 0
		num_pos = 0
		num_nnz = 0
		cfs = Counter()
		dfs = Counter()
		f = gzip.open(self.path+corpus_file, 'rt',encoding='utf8')
		f = tqdm(f,'dictionary',self.sentences_cnt)
		unique = set()
		for line in f:
			line = line.strip()
			if not line: # end of document
				dfs.update(unique)
				num_nnz += len(unique)
				#
				doc_id += 1
				unique = set()
				continue
			tokens = line.split(' ')
			cfs.update(tokens)
			num_pos += len(tokens)
			unique.update(tokens)
		f.close()
		#
		token2id = {t:i for i,(t,cnt) in enumerate(cfs.most_common())}
		dictionary = GensimDictionary()
		dictionary.num_pos = num_pos
		dictionary.num_nnz = num_nnz
		dictionary.num_docs = doc_id
		dictionary.token2id = token2id
		#dictionary.cfs = {i:cfs[t] for t,i in token2id.items()}
		#dictionary.dfs = {i:dfs[t] for t,i in token2id.items()}
		for t,i in token2id.items():
			dictionary.cfs[i] = cfs[t]
			dictionary.dfs[i] = dfs[t]
		#dictionary.patch_with_special_tokens({'<PAD>':0})
		if save:
			dictionary.save(self.path+'dictionary.pkl')
		self.dictionary = dictionary
	
	def load_dictionary(self):
		self.dictionary = GensimDictionary.load(self.path+'dictionary.pkl')

	@timed
	def prune_dictionary(self, stopwords=None, save=True, **kwargs):
		_stopwords = stopwords or self.params.get('dictionary__stopwords')
		self.dictionary.filter_extremes(**kwargs)
		if _stopwords:
			bad_ids = [self.dictionary.token2id.get(t,-1) for t in _stopwords]
			self.dictionary.filter_tokens(bad_ids)
		self.dictionary.compactify()
		if save:
			self.dictionary.save(self.path+'dictionary.pkl')

	@timed
	def init_bow_old(self, storage='disk'):
		self.bow = sorbet(self.path+'bow', kind=storage).new()
		source = self.phrased
		src = tqdm(source, desc='bow', total=len(self.meta))
		for doc in src:
			bow = self.dictionary.doc2bow(doc) or [(0,1)] # 0 -> '<PAD>'
			self.bow.append(bow)
		self.bow.save()

	@timed
	def init_bow(self, storage=None):
		_storage = storage or self.params.get('bow__storage') or self.params.get('storage','disk')
		self.bow = sorbet(self.path+'bow', kind=_storage).new()
		doc = []
		for doc_id,sen in self.all_sentences(desc='bow'):
			if not sen: # end of document:		
				bow = self.dictionary.doc2bow(doc) or [(0,1)] # 0 -> '<PAD>'
				self.bow.append(bow)
				doc = []
			else:
				tokens = sen.split(' ')
				doc.extend(tokens)
		self.bow.save()

	def load_bow(self, storage=None):
		_storage = storage or self.params.get('bow__storage') or self.params.get('storage','disk')
		self.bow = sorbet(self.path+'bow', kind=storage).load()
	
	# TODO opcjonalna inicjalizacja z sentences
	@timed
	def init_inverted(self):
		inverted = {}
		bow = tqdm(self.bow, desc='inverted', total=len(self.bow))
		for i,b in enumerate(bow):
			for t,cnt in b:
				if t in inverted:
					inverted[t] += [i]
				else:
					inverted[t] = [i]
		#
		self.inverted = sorbet(self.path+'inverted').new()
		for t in range(max(inverted)):
			self.inverted.append(set(inverted.get(t,[])))
		self.inverted.save()

	def load_inverted(self):
		self.inverted = sorbet(self.path+'inverted').load()

	# TODO findall vs match
	def dictionary_query(self, query, exclude=None):
		"returns list of ids and list of tokens matching the query"
		q = re.compile(query)
		e = re.compile(exclude) if exclude else None
		ids = []
		tokens = []
		for i,token in self.dictionary.items():
			if q.findall(token):
				if e and e.findall(token): continue
				ids += [i]
				tokens += [token]
		return ids,tokens

	# TODO zliczanie wystapien
	# TODO sortowanie
	def inverted_query(self, query, exclude=None):
		"returns ids of documents containing matching tokens"
		doc_ids = set()
		token_ids,tokens = self.dictionary_query(query, exclude)
		for i in token_ids:
			doc_ids.update(self.inverted[i])
		return doc_ids
	
# ---[ DEBUG ]------------------------------------------------------------------

if __name__=="__main__":
	model = Dictionary()
	model.path = '../model_10/'
	model.load_dictionary()
	#
	d = list(model.dictionary.dfs.items())
	d.sort(key=lambda x:-x[1])
	print(f'NUM_DOCS: {model.dictionary.num_docs}')
	print('\nSINGLE:')
	for id,cnt in d[:60]:
		token = model.dictionary.get(id)
		print(f"{cnt:4d} {token}")
	print('\nPHRASES:')
	i=0
	for id,cnt in d:
		token = model.dictionary.get(id)
		if '__' not in token: continue
		print(f"{cnt:4d} {token}")
		i+=1
		if i>200:break

